@@ -52,7 +52,7 @@ from .utils.gcode import gcode
 from .utils.fluid_analysis import reynolds, wss
 from .utils.gui import gui_helper
 
-from .tree import tree, forest, perfusion_territory
+from .tree import tree, forest, perfusion_territory, get
 from .implicit.implicit import surface
 from .implicit.tests.bumpy_sphere import bumpy_sphere
 from .implicit.tests.heart import *
